; remove unused states from randomly generated ants

; (module remove-ununsed-states	(main main))

(define (main argv)
  (let* ((program (read-program (cadr argv)))
	 (p2 (remove-unused-states program))
	 (st1 (number-of-states program))
	 (st2 (number-of-states p2)))
    (display "number of states in source program: ") (display st1) (newline)
    (display "number of states after reducing ununsed: ") (display st2) (newline)
    (if (< st2 st1)
	(write-program p2 (caddr argv)))))

(define (number-of-states p) (length p))


; instruction:
; (instr arg ...)

; table of instructions and their state args
(define instr-states
  '((sense 2 3)
    (mark 2)
    (nmark 2)
    (pickup 1 2)
    (drop 1)
    (turn 2)
    (move 1 2)
    (flip 2 3)))


(define (read-program file)
  (with-input-from-file file
    (lambda ()
      (letrec ((h (lambda (n r)
		    (let ((line (read-line)))
		      (if (eof-object? line)
			  (reverse r)
			  (h (+ n 1) (cons (cons n (string->state line)) r)))))))
	(h 0 '())))))

(define (string->state s)
  (words s))


(define (words s)
  (letrec ((h (lambda (p ws)
		(let ((w (read p)))
		  (if (eof-object? w)
		      (reverse ws)
		      (h p (cons w ws)))))))
    (h (open-input-string s) '())))


(define (used-states program)
  (letrec ((follow (lambda (prog state used)
		     (let ((instruction (list-ref prog state)))
		       (if (car instruction)
			   used
			   (begin
			     (set-car! instruction #t)
			     (let ((states (cdr (assq (caddr instruction) instr-states)))
				   (used1 (cons state used)))
			       (if (= (length states) 1)
				   (follow prog (list-ref instruction (+ (car states) 2)) used1)
				   (follow prog (list-ref instruction (+ (car states) 2))
					   (follow prog (list-ref instruction (+ (cadr states) 2)) used1))))))))))
    (follow (map (lambda (i) (cons #f i)) program)
	    0
	    '())))


(define (remove-unused-states program)
  (let ((used (used-states program)))
    (assemble (filter (lambda (s) (memv (car s) used)) program))))


(define (assemble is)
  (map (lambda (i) (assemble-instruction i is)) is))

(define (assemble-instruction i is)
  (case (cadr i)
    ((sense)
     (if (eq? (list-ref i 5) 'marker)
	 (list 'sense (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is) (list-ref i 5)
	       (list-ref i 6))
	 (list 'sense (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is) (list-ref i 5))))
    ((mark unmark turn) (list (cadr i) (caddr i) (stateno (cadddr i) is)))
    ((pickup move) (list (cadr i) (stateno (caddr i) is) (stateno (cadddr i) is)))
    ((drop) (list 'drop (stateno (caddr i) is)))
    ((flip) (list 'flip (caddr i) (stateno (cadddr i) is) (stateno (list-ref i 4) is)))
    (else
     (error "assemble-instruction" "unknown instruction" i))))

(define (stateno state is)
  (letrec ((h (lambda (n is)
		(cond ((null? is) (error "stateno" "state not found" state))
		      ((eq? state (caar is)) n)
		      (else (h (+ n 1) (cdr is)))))))
    (h 0 is)))




(define (write-program program file)
  (with-output-to-file file
    (lambda ()
      (write-instructions program))))

(define (write-instructions program)
  (if (null? program)
      #f
      (begin (write-instruction (car program))
	     (newline)
	     (write-instructions (cdr program)))))

(define (write-instruction i)
  (display (car i))
  (do ((args (cdr i) (cdr args)))
      ((null? args))
    (display #\space)
    (display (car args))))


