#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static const char *opponents[] = {
	"./breed_opponent1.ant",
	"./breed_opponent2.ant",
	NULL
};


struct ant_t {
	const char *ant;
	int idx;
	int score;
};

static int cmp_score(const void *a, const void *b)
{
	const struct ant_t *a1 = a;
	const struct ant_t *a2 = b;
	return a1->score - a2->score;
}


static void gen_ant(const char *antfile, const struct ant_t *pool, int pool_size)
{
	char cmd[200];
	int i, j;
	int r = rand() % 16;

	if (r >= 8) {
		/* crossover. */
		i = rand() % pool_size;
		do {
			j = rand() % pool_size;
		} while (j == i);
		printf("replacing %s by crossover of %s and %s.\n",
		       antfile, pool[i].ant, pool[j].ant);
		sprintf(cmd, "../simulator/crossover %s %s 30 > %s",
			pool[i].ant, pool[j].ant, antfile);
		system(cmd);
	}
	else if (r > 0) {
		/* mutate. */
		i = rand() % pool_size;
		printf("replacing %s by mutation of %s.\n", antfile, pool[i].ant);
		sprintf(cmd, "../simulator/ant_mutate %s 20 > %s", pool[i].ant, antfile);
		system(cmd);
	}
	else {
		/* new random ant. */
		printf("replacing %s by new random ant.\n", antfile);
		sprintf(cmd, "../simulator/antgen 10000 > %s", antfile);
		system(cmd);
	}
}


static void contest(const char *world, const char *red, const char *black,
		    int *red_score, int *black_score)
{
	static const char temp[] = "breed.tmp";
	FILE *f;
	char cmd[300];

	sprintf(cmd, "../simulator/simulator_contest %s %s %s 100000 > %s 2>&1",
		world, red, black, temp);
	system(cmd);
	f = fopen(temp, "r");
	if (f == NULL) {
		fprintf(stderr, "Error: can't open %s.\n", temp);
		exit(1);
	}
	if (fscanf(f, "%s", cmd) != 1) {
		fprintf(stderr, "Error: could not parse contest result.\nCommand was: %s\n", cmd);
		exit(1);
	}
	fclose(f);
	remove(temp);
	if (strcmp(cmd, "tie") == 0) {
		*red_score = 1;
		*black_score = 1;
	}
	else if (strcmp(cmd, "Red") == 0) {
		*red_score = 2;
		*black_score = 0;
	}
	else if (strcmp(cmd, "Black") == 0) {
		*red_score = 0;
		*black_score = 2;
	}
	else {
		fprintf(stderr, "Could not parse result: %s.\n", cmd);
		exit(1);
	}
}


static int get_score(const char *ant)
{
	static const char *worlds[] = {
		"../worlds/sample0.world",
		"../worlds/sample1.world",
		"../worlds/sample2.world",
		"../worlds/sample3.world",
		"../worlds/sample4.world",
		"../worlds/sample5.world",
		"../worlds/sample6.world",
		"../worlds/sample7.world",
		"../worlds/sample8.world",
		"../worlds/sample9.world",
		NULL
	};

	int i;
	int score = 0;

	for (i = 0; worlds[i] != NULL; i++) {
		int j;
		for (j = 0; opponents[j] != NULL; j++) {
			int a, b;
			contest(worlds[i], opponents[j], ant, &a, &b);
			score += b;
			contest(worlds[i], ant, opponents[j], &a, &b);
			score += a;
		}
	}
	printf("%s score: %d.\n", ant, score);
	return score;
}



int main(int argc, char **argv)
{
	int generation;
	int num_ants = argc - 1;
	char **ant = argv + 1;
	struct ant_t *score;
	int *last_score;
	int i;

	score = malloc(num_ants * sizeof(struct ant_t));
	last_score = malloc(num_ants * sizeof(int));
	for (i = 0; i < num_ants; i++)
		last_score[i] = -1;

	for (generation = 0;; generation++) {
		FILE *f;
		int low;

		if ((f = fopen("stop_breed", "r")) != NULL) {
			fclose(f);
			remove("stop_breed");
			break;
		}
		printf("Generation %d:\n", generation);
		for (i = 0; i < num_ants; i++) {
			score[i].ant = ant[i];
			score[i].idx = i;
			if (last_score[i] >= 0)
				score[i].score = last_score[i];
			else {
				score[i].score = get_score(ant[i]);
				last_score[i] = score[i].score;
			}
		}

		/* sort ascending. */
		qsort(score, num_ants, sizeof (struct ant_t), cmp_score);
		for (i = num_ants - 1; i >= 0; i--) {
			printf("%4d %s\n", score[i].score, score[i].ant);
		}
		for (i = 0; i < 2; i++) {
			/* swap two random ants. */
			int p = rand() % num_ants;
			int q = rand() % num_ants;
			struct ant_t tmp;
			tmp = score[p];
			score[p] = score[q];
			score[q] = tmp;
		}



		/* regenerate lower scoring ants. */
		low = num_ants / 2;
		for (i = 0; i < low; i++) {
			last_score[score[i].idx] = -1;
			gen_ant(score[i].ant, score + low, num_ants - low);
		}
	}
	return 0;
}
