/* simulator driver program. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "world.h"
#include "rng.h"


int main(int argc, char **argv)
{
	int steps;
	int s;
	struct world *w;
	const struct instruction *ant1, *ant2;
	FILE *f;
	bool dump = false;

	if (dump)
		printf("Simulator.\n");
	if (argc != 5) {
		fprintf(stderr, "Usage: %s world red-ant black-ant steps\n", argv[0]);
		exit(1);
	}

	if (strcmp(argv[1], "../tiny.world") == 0)
		dump = true;

	ant1 = read_program(argv[2]);
	ant2 = read_program(argv[3]);
	w = read_world(argv[1], ant1, ant2);
	steps = atoi(argv[4]);

	if (dump) {
		f = fopen("sim.ant-red.log", "w");
		print_program(ant1, f);
		fclose(f);
		f = fopen("sim.ant-black.log", "w");
		print_program(ant2, f);
		fclose(f);
		f = fopen("sim.dump.log", "w");
	} else {
		f = NULL;
	}

	rng_init();
	for (s = 0; s < steps; s++) {
		int i;
		if (f != NULL) {
			fprintf(f, "\nAfter round %d...\n", s);
			dump_world(w, f);
			/* print_statistics(w, f); */
		}
		for (i = 0; i < w->num_ants; i++) {
			struct ant *a = &w->ants[i];
			if (a->id == i)
				step(w, a);
		}
	}
	print_statistics(w, stdout);
	if (f != NULL) fclose(f);

	return 0;
}
