/* The world. */

#ifndef WORLD_H
#define WORLD_H 1

#ifndef NO_STDBOOL
#include <stdbool.h>
#else
typedef int bool;
#define true 1
#define false 0
#endif
#include <stddef.h>
#include <stdio.h>

#define MAX_INSTRUCTIONS 10000


enum sense_direction {
	Here, Ahead, LeftAhead, RightAhead
};

enum condition {
	Friend, Foe, FriendWithFood, FoeWithFood,
	Food, Rock, Marker, FoeMarker,
	Home, FoeHome
};

enum turn_dir { Left, Right };

struct world;
struct ant;

struct instruction {
	void (*code)(const struct instruction *, struct world *, struct ant *);
	void (*print)(const struct instruction *, FILE *);
	int marker;
	int state;
	int state2;
	enum sense_direction sense_dir;
	enum condition cond;
	enum turn_dir left_or_right;
#ifdef FAST_RNG
	unsigned int rng_mask;
#endif
	unsigned int flipval;
};



enum color { Red, Black };

extern const enum color other_color_x[2];
#define other_color(c) (other_color_x[(c)])

struct ant {
	int id;			/* -1 if dead */
	enum color color;
	int state;
	int resting; /* 0 - 14 - FIXME: or 0 - 13 ??? */
	int direction; /* 0 - 5 */
	bool has_food;
	struct cell *p;
	const struct instruction *program;
};

static inline const struct instruction *get_current_instruction(const struct ant *a)
{
	return &a->program[a->state];
}

extern const int turns[2][6];
#define turn(lr,d) (turns[(lr)][(d)])


struct cell {
	struct ant *ant;	/* NULL if no ant here. */
	unsigned int food;
	char spec;
	unsigned char markers[2];
	struct cell *adj[6];	/* adjacent cells. */
};

#define adjacent_cell(w,c,dir) ((c)->adj[(dir)])

#ifdef CONTEST_SIZE
#define WIDTH(w) 100
#define HEIGHT(w) 100
#else
#define WIDTH(w) ((w)->width)
#define HEIGHT(w) ((w)->height)
#endif

struct world {
	int width;
	int height;
	struct cell *grid;
	struct ant *ants;	/* ants[num_ants] == NULL */
	int num_ants;
};


static inline struct cell *getcell(const struct world *w, int x, int y)
{
	return &w->grid[y * WIDTH(w) + x];
}


static inline void set_marker(struct ant *a, int marker)
{
	a->p->markers[a->color] |= (1 << marker);
}

static inline void clear_marker(struct ant *a, int marker)
{
	a->p->markers[a->color] &= ~(1 << marker);
}


extern struct cell *sensed_cell(const struct world *w, struct cell *c, int dir, enum sense_direction sense_dir);
extern bool cell_matches(const struct cell *c, enum condition cond, int marker, enum color col);


extern void check_for_surrounded_ants(struct world *, struct cell *);

extern struct instruction *read_program(const char *);
extern void print_program(const struct instruction *, FILE *);

extern void dump_world(const struct world *, FILE *);
extern struct world *read_world(const char *, const struct instruction *, const struct instruction *);

extern void step(struct world *, struct ant *);
extern void print_statistics(const struct world *, FILE *);

extern void str_lower(char *);
extern void *xmalloc(size_t) __attribute__((malloc));

extern void random_instruction(struct instruction *, int, int);
extern void random_instruction_weighted(struct instruction *, int, int);

#endif /* WORLD_H */
